/* Style sheet for the AjaxSuggest example. */

.in_red {
	color: #a33;
}

.suggest_button_normal {
	color: #000;
	background-color: #ddd;
	text-align: center;
	font-size: smaller;
	padding: 1px 4px;
	border: 2px outset #666;
}

.suggest_button_over {
	color: #ddd;
	background-color: #36c;
	text-align: center;
	font-size: smaller;
	padding: 1px 4px;
	border: 2px outset #666;
}

.suggest_link_normal {
	color: #000;
	background-color: #fff;
	padding: 2px 4px;
}

.suggest_link_over {
	color: #fff;
	background-color: #36c;
	padding: 2px 4px;
}

#query {
	padding: 2px 4px;
	border: 2px solid #000;
}

#suggestions {
	cursor: pointer;
	position: absolute;
	text-align: left;
	border: 1px solid #666;
}

.show {
	display: block;
}

.hide {
	display: none;
}
