name = 'Python Server Pages'

status = 'stable'

synopsis = (
    "A Python Server Page (or PSP) is an HTML document"
    " with interspersed Python instructions that are interpreted"
    " as a template to generate dynamic content."
    " PSP is analogous to PHP, Microsoft's ASP and Sun's JSP."
    " PSP sits on top of (and requires) Webware and therefore"
    " benefits from its features.")

webwareConfig = {
    'examplePages': [
        'Hello',
        'Braces',
        'PSPTests',
        'PSPTests-Braces',
    ]
}
