name = 'Webware for Python'

version = (3, 0, 0)

status = 'stable'

requiredPyVersion = (3, 6)

synopsis = (
    "Webware for Python is a time-tested"
    " modular, object-oriented web framework.")

webwareConfig = {
    'examplePages': [
        'Welcome',
        'ShowTime',
        'CountVisits',
        'Error',
        'View',
        'Introspect',
        'Colors',
        'ListBox',
        'Forward',
        'SecureCountVisits',
        'FileUpload',
        'RequestInformation',
        'ImageDemo',
        'DominateDemo',
        'YattagDemo',
        'DBUtilsDemo',
        'AjaxSuggest',
        'JSONRPCClient',
    ]
}
